'''///////////////////////////////////////////////////////////////////////////////
'	Wiimote.cs
'	Managed Wiimote Library
'	Written by Brian Peek (http://www.brianpeek.com/)
'	for MSDN's Coding4Fun (http://msdn.microsoft.com/coding4fun/)
'	Visit http://blogs.msdn.com/coding4fun/archive/2007/03/14/1879033.aspx
'	for more information
'''///////////////////////////////////////////////////////////////////////////////

Imports System.Runtime.InteropServices
Imports System.Diagnostics
Imports System.IO
Imports Microsoft.Win32.SafeHandles
Imports System.Threading

Namespace WiimoteLib
    ''' <summary>
    ''' Implementation of Wiimote
    ''' </summary>
    Public Class Wiimote
        Implements IDisposable
        ''' <summary>
        ''' Event raised when Wiimote state is changed
        ''' </summary>
        Public Event WiimoteChanged As WiimoteChangedEventHandler

        ''' <summary>
        ''' Event raised when an extension is inserted or removed
        ''' </summary>
        Public Event WiimoteExtensionChanged As WiimoteExtensionChangedEventHandler

        ' VID = Nintendo, PID = Wiimote
        Private Const VID As Integer = &H57E
        Private Const PID As Integer = &H330 'Wiimote Plus
        Private Const PID2 As Integer = &H306 'Wiimote

        ' sure, we could find this out the hard way using HID, but trust me, it's 22
        Private Const REPORT_LENGTH As Integer = 22

        ''' <summary>
        ''' The report format in which the Wiimote should return data
        ''' </summary>	
        Public Enum InputReport As Byte
            ''' <summary>
            ''' Status report
            ''' </summary>
            Status = &H20
            ''' <summary>
            ''' Read data from memory location
            ''' </summary>
            ReadData = &H21
            ''' <summary>
            ''' Button data only
            ''' </summary>
            Buttons = &H30
            ''' <summary>
            ''' Button and accelerometer data
            ''' </summary>
            ButtonsAccel = &H31
            ''' <summary>
            ''' IR sensor and accelerometer data
            ''' </summary>
            IRAccel = &H33
            ''' <summary>
            ''' Button and extension controller data
            ''' </summary>
            ButtonsExtension = &H34
            ''' <summary>
            ''' Extension and accelerometer data
            ''' </summary>
            ExtensionAccel = &H35
            ''' <summary>
            ''' IR sensor, extension controller and accelerometer data
            ''' </summary>
            IRExtensionAccel = &H37
        End Enum

        ' Wiimote output commands
        Private Enum OutputReport As Byte
            None = &H0
            LEDs = &H11
            Type = &H12
            IR = &H13
            Status = &H15
            WriteMemory = &H16
            ReadMemory = &H17
            IR2 = &H1A
        End Enum

        ' Wiimote registers
        Private Const REGISTER_IR As Integer = &H4B00030
        Private Const REGISTER_IR_SENSITIVITY_1 As Integer = &H4B00000
        Private Const REGISTER_IR_SENSITIVITY_2 As Integer = &H4B0001A
        Private Const REGISTER_IR_MODE As Integer = &H4B00033

        Private Const REGISTER_EXTENSION_INIT As Integer = &H4A40040
        Private Const REGISTER_EXTENSION_TYPE As Integer = &H4A400FE
        Private Const REGISTER_EXTENSION_CALIBRATION As Integer = &H4A40020

        ' read/write handle to the device
        Private mHandle As SafeFileHandle

        ' a pretty .NET stream to read/write from/to
        Private mStream As FileStream

        ' report buffer
        Private ReadOnly mBuff As Byte() = New Byte(REPORT_LENGTH - 1) {}

        ' read data buffer
        Private mReadBuff As Byte()

        ' current state of controller
        Private ReadOnly mWiimoteState As New WiimoteState()

        ' event for read data processing
        Private ReadOnly mReadDone As New AutoResetEvent(False)

        ' use a different method to write reports
        Private mAltWriteMethod As Boolean

        ''' <summary>
        ''' Default constructor
        ''' </summary>
        Public Sub New()
        End Sub

        ''' <summary>
        ''' Connect to a Wiimote paired to the PC via Bluetooth
        ''' </summary>
        Public Sub Connect()
            Dim index As Integer = 0
            Dim found As Boolean = False
            Dim guid As Guid

            ' get the GUID of the HID class
            HIDImports.HidD_GetHidGuid(guid)

            ' get a handle to all devices that are part of the HID class
            ' Fun fact:  DIGCF_PRESENT worked on my machine just fine.  I reinstalled Vista, and now it no longer finds the Wiimote with that parameter enabled...
            Dim hDevInfo As IntPtr = HIDImports.SetupDiGetClassDevs(guid, Nothing, IntPtr.Zero, HIDImports.DIGCF_DEVICEINTERFACE)
            ' | HIDImports.DIGCF_PRESENT);
            ' create a new interface data struct and initialize its size
            Dim diData As New HIDImports.SP_DEVICE_INTERFACE_DATA()
            diData.cbSize = Marshal.SizeOf(diData)

            ' get a device interface to a single device (enumerate all devices)
            While HIDImports.SetupDiEnumDeviceInterfaces(hDevInfo, IntPtr.Zero, guid, index, diData)
                Dim size As UInt32

                ' get the buffer size for this device detail instance (returned in the size parameter)
                HIDImports.SetupDiGetDeviceInterfaceDetail(hDevInfo, diData, IntPtr.Zero, 0, size, IntPtr.Zero)

                ' create a detail struct and set its size
                Dim diDetail As New HIDImports.SP_DEVICE_INTERFACE_DETAIL_DATA()

                ' yeah, yeah...well, see, on Win x86, cbSize must be 5 for some reason.  On x64, apparently 8 is what it wants.
                ' someday I should figure this out.  Thanks to Paul Miller on this...
                If IntPtr.Size = 8 Then
                    diDetail.cbSize = 8
                Else
                    diDetail.cbSize = 5
                End If

                ' actually get the detail struct
                If HIDImports.SetupDiGetDeviceInterfaceDetail(hDevInfo, diData, diDetail, size, size, IntPtr.Zero) Then
                    Debug.WriteLine((index & " ") + diDetail.DevicePath & " " & Marshal.GetLastWin32Error())

                    ' open a read/write handle to our device using the DevicePath returned
                    mHandle = HIDImports.CreateFile(diDetail.DevicePath, FileAccess.ReadWrite, FileShare.ReadWrite, IntPtr.Zero, FileMode.Open, HIDImports.EFileAttributes.Overlapped, _
                     IntPtr.Zero)

                    ' create an attributes struct and initialize the size
                    Dim attrib As New HIDImports.HIDD_ATTRIBUTES()
                    attrib.Size = Marshal.SizeOf(attrib)

                    ' get the attributes of the current device
                    If HIDImports.HidD_GetAttributes(mHandle.DangerousGetHandle(), attrib) Then
                        ' if the vendor and product IDs match up
                        If attrib.VendorID = VID AndAlso attrib.ProductID = PID Or attrib.ProductID = PID2 Then
                            Debug.WriteLine("Found it!")
                            found = True

                            ' create a nice .NET FileStream wrapping the handle above
                            mStream = New FileStream(mHandle, FileAccess.ReadWrite, REPORT_LENGTH, True)

                            ' start an async read operation on it
                            BeginAsyncRead()

                            ' read the calibration info from the controller
                            Try
                                ReadCalibration()
                            Catch
                                ' if we fail above, try the alternate HID writes
                                mAltWriteMethod = True
                                ReadCalibration()
                            End Try

                            ' force a status check to get the state of any extensions plugged in at startup
                            GetStatus()

                            Exit While
                        Else
                            ' otherwise this isn't the controller, so close up the file handle
                            mHandle.Close()
                        End If
                    End If
                Else
                    ' failed to get the detail struct
                    Throw New Exception("SetupDiGetDeviceInterfaceDetail failed on index " & index)
                End If

                ' move to the next device
                index += 1
            End While

            ' clean up our list
            HIDImports.SetupDiDestroyDeviceInfoList(hDevInfo)

            ' if we didn't find a Wiimote, throw an exception
            If Not found Then
                Throw New Exception("Wiimote not found in HID device list.")
            End If
        End Sub

        ''' <summary>
        ''' Disconnect from the controller and stop reading data from it
        ''' </summary>
        Public Sub Disconnect()
            ' close up the stream and handle
            If mStream IsNot Nothing Then
                mStream.Close()
            End If

            If mHandle IsNot Nothing Then
                mHandle.Close()
            End If
        End Sub

        ''' <summary>
        ''' Start reading asynchronously from the controller
        ''' </summary>
        Private Sub BeginAsyncRead()
            ' if the stream is valid and ready
            If mStream.CanRead Then
                ' setup the read and the callback
                Dim buff As Byte() = New Byte(REPORT_LENGTH - 1) {}
                mStream.BeginRead(buff, 0, REPORT_LENGTH, New AsyncCallback(AddressOf OnReadData), buff)
            End If
        End Sub

        ''' <summary>
        ''' Callback when data is ready to be processed
        ''' </summary>
        ''' <param name="ar">State information for the callback</param>
        Private Sub OnReadData(ByVal ar As IAsyncResult)
            ' grab the byte buffer
            Dim buff As Byte() = DirectCast(ar.AsyncState, Byte())

            Try
                ' end the current read
                mStream.EndRead(ar)

                ' parse it
                If ParseInput(buff) Then
                    ' post an event
                    RaiseEvent WiimoteChanged(Me, New WiimoteChangedEventArgs(mWiimoteState))
                End If

                ' start reading again
                BeginAsyncRead()
            Catch generatedExceptionName As OperationCanceledException
                Debug.WriteLine("OperationCanceledException")
            End Try
        End Sub

        ''' <summary>
        ''' Parse a report sent by the Wiimote
        ''' </summary>
        ''' <param name="buff">Data buffer to parse</param>
        ''' <returns>Returns a boolean noting whether an event needs to be posted</returns>
        Private Function ParseInput(ByVal buff As Byte()) As Boolean
            Dim type As InputReport = CType(buff(0), InputReport)

            Select Case type
                Case InputReport.Buttons
                    ParseButtons(buff)
                    Exit Select
                Case InputReport.ButtonsAccel
                    ParseButtons(buff)
                    ParseAccel(buff)
                    Exit Select
                Case InputReport.IRAccel
                    ParseButtons(buff)
                    ParseAccel(buff)
                    ParseIR(buff)
                    Exit Select
                Case InputReport.ButtonsExtension
                    ParseButtons(buff)
                    ParseExtension(DecryptBuffer(buff), 4)
                    Exit Select
                Case InputReport.ExtensionAccel
                    ParseButtons(buff)
                    ParseAccel(buff)
                    ParseExtension(DecryptBuffer(buff), 6)
                    Exit Select
                Case InputReport.IRExtensionAccel
                    ParseButtons(buff)
                    ParseAccel(buff)
                    ParseIR(buff)
                    ParseExtension(DecryptBuffer(buff), 16)
                    Exit Select
                Case InputReport.Status
                    ParseButtons(buff)
                    mWiimoteState.Battery = buff(6)

                    ' get the real LED values in case the values from SetLEDs() somehow becomes out of sync, which really shouldn't be possible
                    mWiimoteState.LEDState.LED1 = (buff(3) And &H10) <> 0
                    mWiimoteState.LEDState.LED2 = (buff(3) And &H20) <> 0
                    mWiimoteState.LEDState.LED3 = (buff(3) And &H40) <> 0
                    mWiimoteState.LEDState.LED4 = (buff(3) And &H80) <> 0

                    ' extension connected?
                    Dim extension As Boolean = (buff(3) And &H2) <> 0
                    Debug.WriteLine("Extension: " & extension)

                    If mWiimoteState.Extension <> extension Then
                        mWiimoteState.Extension = extension

                        If extension Then
                            ' start reading again
                            BeginAsyncRead()

                            InitializeExtension()
                        Else
                            mWiimoteState.ExtensionType = ExtensionType.None
                        End If

                        RaiseEvent WiimoteExtensionChanged(Me, New WiimoteExtensionChangedEventArgs(mWiimoteState.ExtensionType, mWiimoteState.Extension))
                    End If

                    Exit Select
                Case InputReport.ReadData
                    ParseButtons(buff)
                    ParseReadData(buff)
                    Exit Select
                Case Else
                    Debug.WriteLine("Unknown report type: " & type.ToString("x"))
                    Return False
            End Select

            Return True
        End Function

        ''' <summary>
        ''' Handles setting up an extension when plugged.  Currenlty only support the Nunchuk.
        ''' </summary>
        Private Sub InitializeExtension()
            WriteData(REGISTER_EXTENSION_INIT, &H0)

            Dim buff As Byte() = ReadData(REGISTER_EXTENSION_TYPE, 2)

            If buff(0) = CByte(ExtensionType.Nunchuk) AndAlso buff(1) = CByte(ExtensionType.Nunchuk) Then
                mWiimoteState.ExtensionType = ExtensionType.Nunchuk
            ElseIf buff(0) = CByte(ExtensionType.ClassicController) AndAlso buff(1) = CByte(ExtensionType.ClassicController) Then
                mWiimoteState.ExtensionType = ExtensionType.ClassicController
            ElseIf buff(0) = &HFF Then
                ' partially inserted case...reset back to nothing inserted
                mWiimoteState.Extension = False
                mWiimoteState.ExtensionType = ExtensionType.None
                Return
            Else
                Throw New Exception("Unknown extension controller found: " & buff(0))
            End If

            buff = DecryptBuffer(ReadData(REGISTER_EXTENSION_CALIBRATION, 16))

            Select Case mWiimoteState.ExtensionType
                Case ExtensionType.Nunchuk
                    mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.X0 = buff(0)
                    mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.Y0 = buff(1)
                    mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.Z0 = buff(2)
                    mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.XG = buff(4)
                    mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.YG = buff(5)
                    mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.ZG = buff(6)
                    mWiimoteState.NunchukState.CalibrationInfo.MaxX = buff(8)
                    mWiimoteState.NunchukState.CalibrationInfo.MinX = buff(9)
                    mWiimoteState.NunchukState.CalibrationInfo.MidX = buff(10)
                    mWiimoteState.NunchukState.CalibrationInfo.MaxY = buff(11)
                    mWiimoteState.NunchukState.CalibrationInfo.MinY = buff(12)
                    mWiimoteState.NunchukState.CalibrationInfo.MidY = buff(13)
                    Exit Select
                Case ExtensionType.ClassicController
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MaxXL = CByte(buff(0) >> 2)
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MinXL = CByte(buff(1) >> 2)
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MidXL = CByte(buff(2) >> 2)
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MaxYL = CByte(buff(3) >> 2)
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MinYL = CByte(buff(4) >> 2)
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MidYL = CByte(buff(5) >> 2)

                    mWiimoteState.ClassicControllerState.CalibrationInfo.MaxXR = CByte(buff(6) >> 3)
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MinXR = CByte(buff(7) >> 3)
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MidXR = CByte(buff(8) >> 3)
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MaxYR = CByte(buff(9) >> 3)
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MinYR = CByte(buff(10) >> 3)
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MidYR = CByte(buff(11) >> 3)

                    ' this doesn't seem right...
                    '					mWiimoteState.ClassicControllerState.AccelCalibrationInfo.MinTriggerL = (byte)(buff[12] >> 3);
                    '					mWiimoteState.ClassicControllerState.AccelCalibrationInfo.MaxTriggerL = (byte)(buff[14] >> 3);
                    '					mWiimoteState.ClassicControllerState.AccelCalibrationInfo.MinTriggerR = (byte)(buff[13] >> 3);
                    '					mWiimoteState.ClassicControllerState.AccelCalibrationInfo.MaxTriggerR = (byte)(buff[15] >> 3);
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MinTriggerL = 0
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MaxTriggerL = 31
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MinTriggerR = 0
                    mWiimoteState.ClassicControllerState.CalibrationInfo.MaxTriggerR = 31
                    Exit Select
            End Select
        End Sub

        ''' <summary>
        ''' Decrypts data sent from the extension to the Wiimote
        ''' </summary>
        ''' <param name="buff">Data buffer</param>
        ''' <returns>Byte array containing decoded data</returns>
        Private Function DecryptBuffer(ByVal buff As Byte()) As Byte()
            For i As Integer = 0 To buff.Length - 1
                buff(i) = CByte(((buff(i) Xor &H17) + &H17) And &HFF)
            Next

            Return buff
        End Function

        ''' <summary>
        ''' Parses a standard button report into the ButtonState struct
        ''' </summary>
        ''' <param name="buff">Data buffer</param>
        Private Sub ParseButtons(ByVal buff As Byte())
            mWiimoteState.ButtonState.A = (buff(2) And &H8) <> 0
            mWiimoteState.ButtonState.B = (buff(2) And &H4) <> 0
            mWiimoteState.ButtonState.Minus = (buff(2) And &H10) <> 0
            mWiimoteState.ButtonState.Home = (buff(2) And &H80) <> 0
            mWiimoteState.ButtonState.Plus = (buff(1) And &H10) <> 0
            mWiimoteState.ButtonState.One = (buff(2) And &H2) <> 0
            mWiimoteState.ButtonState.Two = (buff(2) And &H1) <> 0
            mWiimoteState.ButtonState.Up = (buff(1) And &H8) <> 0
            mWiimoteState.ButtonState.Down = (buff(1) And &H4) <> 0
            mWiimoteState.ButtonState.Left = (buff(1) And &H1) <> 0
            mWiimoteState.ButtonState.Right = (buff(1) And &H2) <> 0
        End Sub

        ''' <summary>
        ''' Parse accelerometer data
        ''' </summary>
        ''' <param name="buff">Data buffer</param>
        Private Sub ParseAccel(ByVal buff As Byte())
            mWiimoteState.AccelState.RawX = buff(3)
            mWiimoteState.AccelState.RawY = buff(4)
            mWiimoteState.AccelState.RawZ = buff(5)

            mWiimoteState.AccelState.X = CSng(CSng(mWiimoteState.AccelState.RawX) - mWiimoteState.AccelCalibrationInfo.X0) / (CSng(mWiimoteState.AccelCalibrationInfo.XG) - mWiimoteState.AccelCalibrationInfo.X0)
            mWiimoteState.AccelState.Y = CSng(CSng(mWiimoteState.AccelState.RawY) - mWiimoteState.AccelCalibrationInfo.Y0) / (CSng(mWiimoteState.AccelCalibrationInfo.YG) - mWiimoteState.AccelCalibrationInfo.Y0)
            mWiimoteState.AccelState.Z = CSng(CSng(mWiimoteState.AccelState.RawZ) - mWiimoteState.AccelCalibrationInfo.Z0) / (CSng(mWiimoteState.AccelCalibrationInfo.ZG) - mWiimoteState.AccelCalibrationInfo.Z0)
        End Sub

        ''' <summary>
        ''' Parse IR data from report
        ''' </summary>
        ''' <param name="buff">Data buffer</param>
        Private Sub ParseIR(ByVal buff As Byte())
            mWiimoteState.IRState.RawX1 = buff(6) Or ((buff(8) >> 4) And &H3) << 8
            mWiimoteState.IRState.RawY1 = buff(7) Or ((buff(8) >> 6) And &H3) << 8

            Select Case mWiimoteState.IRState.Mode
                Case IRMode.Basic
                    mWiimoteState.IRState.RawX2 = buff(9) Or ((buff(8) >> 0) And &H3) << 8
                    mWiimoteState.IRState.RawY2 = buff(10) Or ((buff(8) >> 2) And &H3) << 8

                    mWiimoteState.IRState.Size1 = &H0
                    mWiimoteState.IRState.Size2 = &H0

                    mWiimoteState.IRState.Found1 = Not (buff(6) = &HFF AndAlso buff(7) = &HFF)
                    mWiimoteState.IRState.Found2 = Not (buff(9) = &HFF AndAlso buff(10) = &HFF)




                    mWiimoteState.IRState.RawX3 = buff(12) Or ((buff(14) >> 4) And &H3) << 8
                    mWiimoteState.IRState.RawY3 = buff(13) Or ((buff(14) >> 6) And &H3) << 8
                    mWiimoteState.IRState.RawX4 = buff(15) Or ((buff(17) >> 4) And &H3) << 8
                    mWiimoteState.IRState.RawY4 = buff(16) Or ((buff(17) >> 6) And &H3) << 8

                    mWiimoteState.IRState.Size3 = buff(14) And &HF
                    mWiimoteState.IRState.Size4 = buff(17) And &HF

                    mWiimoteState.IRState.Found3 = Not (buff(12) = &HFF AndAlso buff(13) = &HFF AndAlso buff(14) = &HFF)
                    mWiimoteState.IRState.Found4 = Not (buff(15) = &HFF AndAlso buff(16) = &HFF AndAlso buff(17) = &HFF)

                    Exit Select
                Case IRMode.Full
                    mWiimoteState.IRState.RawX2 = buff(9) Or ((buff(11) >> 4) And &H3) << 8
                    mWiimoteState.IRState.RawY2 = buff(10) Or ((buff(11) >> 6) And &H3) << 8
                    mWiimoteState.IRState.RawX3 = buff(12) Or ((buff(14) >> 4) And &H3) << 8
                    mWiimoteState.IRState.RawY3 = buff(13) Or ((buff(14) >> 6) And &H3) << 8
                    mWiimoteState.IRState.RawX4 = buff(15) Or ((buff(17) >> 4) And &H3) << 8
                    mWiimoteState.IRState.RawY4 = buff(16) Or ((buff(17) >> 6) And &H3) << 8

                    mWiimoteState.IRState.Size1 = buff(8) And &HF
                    mWiimoteState.IRState.Size2 = buff(11) And &HF
                    mWiimoteState.IRState.Size3 = buff(14) And &HF
                    mWiimoteState.IRState.Size4 = buff(17) And &HF

                    mWiimoteState.IRState.Found1 = Not (buff(6) = &HFF AndAlso buff(7) = &HFF AndAlso buff(8) = &HFF)
                    mWiimoteState.IRState.Found2 = Not (buff(9) = &HFF AndAlso buff(10) = &HFF AndAlso buff(11) = &HFF)
                    mWiimoteState.IRState.Found3 = Not (buff(12) = &HFF AndAlso buff(13) = &HFF AndAlso buff(14) = &HFF)
                    mWiimoteState.IRState.Found4 = Not (buff(15) = &HFF AndAlso buff(16) = &HFF AndAlso buff(17) = &HFF)

                    Exit Select
            End Select

            mWiimoteState.IRState.X1 = CSng(mWiimoteState.IRState.RawX1 / 1023.5F)
            mWiimoteState.IRState.X2 = CSng(mWiimoteState.IRState.RawX2 / 1023.5F)
            mWiimoteState.IRState.X3 = CSng(mWiimoteState.IRState.RawX3 / 1023.5F)
            mWiimoteState.IRState.X4 = CSng(mWiimoteState.IRState.RawX4 / 1023.5F)

            mWiimoteState.IRState.Y1 = CSng(mWiimoteState.IRState.RawY1 / 767.5F)
            mWiimoteState.IRState.Y2 = CSng(mWiimoteState.IRState.RawY2 / 767.5F)
            mWiimoteState.IRState.Y3 = CSng(mWiimoteState.IRState.RawY3 / 767.5F)
            mWiimoteState.IRState.Y4 = CSng(mWiimoteState.IRState.RawY4 / 767.5F)

            If mWiimoteState.IRState.Found1 AndAlso mWiimoteState.IRState.Found2 Then
                mWiimoteState.IRState.RawMidX = (mWiimoteState.IRState.RawX2 + mWiimoteState.IRState.RawX1) / 2
                mWiimoteState.IRState.RawMidY = (mWiimoteState.IRState.RawY2 + mWiimoteState.IRState.RawY1) / 2

                mWiimoteState.IRState.MidX = (mWiimoteState.IRState.X2 + mWiimoteState.IRState.X1) / 2.0F
                mWiimoteState.IRState.MidY = (mWiimoteState.IRState.Y2 + mWiimoteState.IRState.Y1) / 2.0F
            Else
                mWiimoteState.IRState.MidX = InlineAssignHelper(mWiimoteState.IRState.MidY, 0.0F)
            End If
        End Sub

        ''' <summary>
        ''' Parse data from an extension.  Nunchuk/Classic Controller support only.
        ''' </summary>
        ''' <param name="buff">Data buffer</param>
        ''' <param name="offset">Offset into data buffer</param>
        Private Sub ParseExtension(ByVal buff As Byte(), ByVal offset As Integer)
            Select Case mWiimoteState.ExtensionType
                Case ExtensionType.Nunchuk
                    mWiimoteState.NunchukState.RawX = buff(offset)
                    mWiimoteState.NunchukState.RawY = buff(offset + 1)
                    mWiimoteState.NunchukState.AccelState.RawX = buff(offset + 2)
                    mWiimoteState.NunchukState.AccelState.RawY = buff(offset + 3)
                    mWiimoteState.NunchukState.AccelState.RawZ = buff(offset + 4)

                    mWiimoteState.NunchukState.C = (buff(offset + 5) And &H2) = 0
                    mWiimoteState.NunchukState.Z = (buff(offset + 5) And &H1) = 0

                    mWiimoteState.NunchukState.AccelState.X = CSng(CSng(mWiimoteState.NunchukState.AccelState.RawX) - mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.X0) / (CSng(mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.XG) - mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.X0)
                    mWiimoteState.NunchukState.AccelState.Y = CSng(CSng(mWiimoteState.NunchukState.AccelState.RawY) - mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.Y0) / (CSng(mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.YG) - mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.Y0)
                    mWiimoteState.NunchukState.AccelState.Z = CSng(CSng(mWiimoteState.NunchukState.AccelState.RawZ) - mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.Z0) / (CSng(mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.ZG) - mWiimoteState.NunchukState.CalibrationInfo.AccelCalibration.Z0)

                    If mWiimoteState.NunchukState.CalibrationInfo.MaxX <> &H0 Then
                        mWiimoteState.NunchukState.X = CSng(CSng(mWiimoteState.NunchukState.RawX) - mWiimoteState.NunchukState.CalibrationInfo.MidX) / (CSng(mWiimoteState.NunchukState.CalibrationInfo.MaxX) - mWiimoteState.NunchukState.CalibrationInfo.MinX)
                    End If

                    If mWiimoteState.NunchukState.CalibrationInfo.MaxY <> &H0 Then
                        mWiimoteState.NunchukState.Y = CSng(CSng(mWiimoteState.NunchukState.RawY) - mWiimoteState.NunchukState.CalibrationInfo.MidY) / (CSng(mWiimoteState.NunchukState.CalibrationInfo.MaxY) - mWiimoteState.NunchukState.CalibrationInfo.MinY)
                    End If

                    Exit Select
                Case ExtensionType.ClassicController
                    mWiimoteState.ClassicControllerState.RawXL = CByte(buff(offset) And &H3F)
                    mWiimoteState.ClassicControllerState.RawYL = CByte(buff(offset + 1) And &H3F)
                    mWiimoteState.ClassicControllerState.RawXR = CByte((buff(offset + 2) >> 7) Or (buff(offset + 1) And &HC0) >> 5 Or (buff(offset) And &HC0) >> 3)
                    mWiimoteState.ClassicControllerState.RawYR = CByte(buff(offset + 2) And &H1F)

                    mWiimoteState.ClassicControllerState.RawTriggerL = CByte(((buff(offset + 2) And &H60) >> 2) Or (buff(offset + 3) >> 5))
                    mWiimoteState.ClassicControllerState.RawTriggerR = CByte(buff(offset + 3) And &H1F)

                    mWiimoteState.ClassicControllerState.ButtonState.TriggerR = (buff(offset + 4) And &H2) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.Plus = (buff(offset + 4) And &H4) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.Home = (buff(offset + 4) And &H8) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.Minus = (buff(offset + 4) And &H10) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.TriggerL = (buff(offset + 4) And &H20) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.Down = (buff(offset + 4) And &H40) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.Right = (buff(offset + 4) And &H80) = 0

                    mWiimoteState.ClassicControllerState.ButtonState.Up = (buff(offset + 5) And &H1) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.Left = (buff(offset + 5) And &H2) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.ZR = (buff(offset + 5) And &H4) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.X = (buff(offset + 5) And &H8) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.A = (buff(offset + 5) And &H10) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.Y = (buff(offset + 5) And &H20) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.B = (buff(offset + 5) And &H40) = 0
                    mWiimoteState.ClassicControllerState.ButtonState.ZL = (buff(offset + 5) And &H80) = 0

                    If mWiimoteState.ClassicControllerState.CalibrationInfo.MaxXL <> &H0 Then
                        mWiimoteState.ClassicControllerState.XL = CSng(CSng(mWiimoteState.ClassicControllerState.RawXL) - mWiimoteState.ClassicControllerState.CalibrationInfo.MidXL) / CSng(mWiimoteState.ClassicControllerState.CalibrationInfo.MaxXL - mWiimoteState.ClassicControllerState.CalibrationInfo.MinXL)
                    End If

                    If mWiimoteState.ClassicControllerState.CalibrationInfo.MaxYL <> &H0 Then
                        mWiimoteState.ClassicControllerState.YL = CSng(CSng(mWiimoteState.ClassicControllerState.RawYL) - mWiimoteState.ClassicControllerState.CalibrationInfo.MidYL) / CSng(mWiimoteState.ClassicControllerState.CalibrationInfo.MaxYL - mWiimoteState.ClassicControllerState.CalibrationInfo.MinYL)
                    End If

                    If mWiimoteState.ClassicControllerState.CalibrationInfo.MaxXR <> &H0 Then
                        mWiimoteState.ClassicControllerState.XR = CSng(CSng(mWiimoteState.ClassicControllerState.RawXR) - mWiimoteState.ClassicControllerState.CalibrationInfo.MidXR) / CSng(mWiimoteState.ClassicControllerState.CalibrationInfo.MaxXR - mWiimoteState.ClassicControllerState.CalibrationInfo.MinXR)
                    End If

                    If mWiimoteState.ClassicControllerState.CalibrationInfo.MaxYR <> &H0 Then
                        mWiimoteState.ClassicControllerState.YR = CSng(CSng(mWiimoteState.ClassicControllerState.RawYR) - mWiimoteState.ClassicControllerState.CalibrationInfo.MidYR) / CSng(mWiimoteState.ClassicControllerState.CalibrationInfo.MaxYR - mWiimoteState.ClassicControllerState.CalibrationInfo.MinYR)
                    End If

                    If mWiimoteState.ClassicControllerState.CalibrationInfo.MaxTriggerL <> &H0 Then
                        mWiimoteState.ClassicControllerState.TriggerL = (mWiimoteState.ClassicControllerState.RawTriggerL) / CSng(mWiimoteState.ClassicControllerState.CalibrationInfo.MaxTriggerL - mWiimoteState.ClassicControllerState.CalibrationInfo.MinTriggerL)
                    End If

                    If mWiimoteState.ClassicControllerState.CalibrationInfo.MaxTriggerR <> &H0 Then
                        mWiimoteState.ClassicControllerState.TriggerR = (mWiimoteState.ClassicControllerState.RawTriggerR) / CSng(mWiimoteState.ClassicControllerState.CalibrationInfo.MaxTriggerR - mWiimoteState.ClassicControllerState.CalibrationInfo.MinTriggerR)
                    End If

                    Exit Select
            End Select
        End Sub

        ''' <summary>
        ''' Parse data returned from a read report
        ''' </summary>
        ''' <param name="buff">Data buffer</param>
        Private Sub ParseReadData(ByVal buff As Byte())
            If (buff(3) And &H8) <> 0 Then
                Throw New Exception("Error reading data from Wiimote: Bytes do not exist.")
            ElseIf (buff(3) And &H7) <> 0 Then
                Throw New Exception("Error reading data from Wiimote: Attempt to read from write-only registers.")
            Else
                Dim size As Integer = buff(3) >> 4
                Array.Copy(buff, 6, mReadBuff, 0, size + 1)
            End If

            ' set the event so the other thread will continue
            mReadDone.[Set]()
        End Sub

        ''' <summary>
        ''' Returns whether rumble is currently enabled.
        ''' </summary>
        ''' <returns>Byte indicating true (0x01) or false (0x00)</returns>
        Private Function GetRumbleBit() As Byte
            Return CByte(If(mWiimoteState.Rumble, &H1, &H0))
        End Function

        ''' <summary>
        ''' Read calibration information stored on Wiimote
        ''' </summary>
        Private Sub ReadCalibration()
            ' this appears to change the report type to 0x31
            Dim buff As Byte() = ReadData(&H16, 7)

            mWiimoteState.AccelCalibrationInfo.X0 = buff(0)
            mWiimoteState.AccelCalibrationInfo.Y0 = buff(1)
            mWiimoteState.AccelCalibrationInfo.Z0 = buff(2)
            mWiimoteState.AccelCalibrationInfo.XG = buff(4)
            mWiimoteState.AccelCalibrationInfo.YG = buff(5)
            mWiimoteState.AccelCalibrationInfo.ZG = buff(6)
        End Sub

        ''' <summary>
        ''' Set Wiimote reporting mode
        ''' </summary>
        ''' <param name="type">Report type</param>
        ''' <param name="continuous">Continuous data</param>
        Public Sub SetReportType(ByVal type As InputReport, ByVal continuous As Boolean)
            Select Case type
                Case InputReport.IRAccel
                    EnableIR(IRMode.Basic)
                    Exit Select
                Case InputReport.IRExtensionAccel
                    EnableIR(IRMode.Extended)
                    Exit Select
                Case Else
                    DisableIR()
                    Exit Select
            End Select

            ClearReport()
            mBuff(0) = CByte(OutputReport.Type)
            mBuff(1) = CByte((If(continuous, &H4, &H0)) Or CByte(If(mWiimoteState.Rumble, &H1, &H0)))
            mBuff(2) = CByte(type)

            WriteReport()
        End Sub

        ''' <summary>
        ''' Set the LEDs on the Wiimote
        ''' </summary>
        ''' <param name="led1">LED 1</param>
        ''' <param name="led2">LED 2</param>
        ''' <param name="led3">LED 3</param>
        ''' <param name="led4">LED 4</param>
        Public Sub SetLEDs(ByVal led1 As Boolean, ByVal led2 As Boolean, ByVal led3 As Boolean, ByVal led4 As Boolean)
            mWiimoteState.LEDState.LED1 = led1
            mWiimoteState.LEDState.LED2 = led2
            mWiimoteState.LEDState.LED3 = led3
            mWiimoteState.LEDState.LED4 = led4

            ClearReport()

            mBuff(0) = CByte(OutputReport.LEDs)
            mBuff(1) = CByte((If(led1, &H10, &H0)) Or (If(led2, &H20, &H0)) Or (If(led3, &H40, &H0)) Or (If(led4, &H80, &H0)) Or GetRumbleBit())

            WriteReport()
        End Sub

        ''' <summary>
        ''' Set the LEDs on the Wiimote
        ''' </summary>
        ''' <param name="leds">The value to be lit up in base2 on the Wiimote</param>
        Public Sub SetLEDs(ByVal leds As Integer)
            mWiimoteState.LEDState.LED1 = (leds And &H1) > 0
            mWiimoteState.LEDState.LED2 = (leds And &H2) > 0
            mWiimoteState.LEDState.LED3 = (leds And &H4) > 0
            mWiimoteState.LEDState.LED4 = (leds And &H8) > 0

            ClearReport()

            mBuff(0) = CByte(OutputReport.LEDs)
            mBuff(1) = CByte((If((leds And &H1) > 0, &H10, &H0)) Or (If((leds And &H2) > 0, &H20, &H0)) Or (If((leds And &H4) > 0, &H40, &H0)) Or (If((leds And &H8) > 0, &H80, &H0)) Or GetRumbleBit())

            WriteReport()
        End Sub

        ''' <summary>
        ''' Toggle rumble
        ''' </summary>
        ''' <param name="on">On or off</param>
        Public Sub SetRumble(ByVal [on] As Boolean)
            mWiimoteState.Rumble = [on]

            ' the LED report also handles rumble
            SetLEDs(mWiimoteState.LEDState.LED1, mWiimoteState.LEDState.LED2, mWiimoteState.LEDState.LED3, mWiimoteState.LEDState.LED4)
        End Sub

        ''' <summary>
        ''' Retrieve the current status of the Wiimote and extensions.  Replaces GetBatteryLevel() since it was poorly named.
        ''' </summary>
        Public Sub GetStatus()
            ClearReport()

            mBuff(0) = CByte(OutputReport.Status)
            mBuff(1) = GetRumbleBit()

            WriteReport()
        End Sub

        ''' <summary>
        ''' Turn on the IR sensor
        ''' </summary>
        ''' <param name="mode">The data report mode</param>
        Private Sub EnableIR(ByVal mode As IRMode)
            mWiimoteState.IRState.Mode = mode

            ClearReport()
            mBuff(0) = CByte(OutputReport.IR)
            mBuff(1) = CByte(&H4 Or GetRumbleBit())
            WriteReport()

            ClearReport()
            mBuff(0) = CByte(OutputReport.IR2)
            mBuff(1) = CByte(&H4 Or GetRumbleBit())
            WriteReport()

            WriteData(REGISTER_IR, &H8)
            'default
            '			WriteData(REGISTER_IR_SENSITIVITY_1, 9, new byte[] {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x90, 0x00, 0xc0});
            '			WriteData(REGISTER_IR_SENSITIVITY_2, 2, new byte[] {0x40, 0x00});

            'super sensitive according to http://wiibrew.org/index.php?title=Wiimote#Sensitivity_Settings
            WriteData(REGISTER_IR_SENSITIVITY_1, 9, New Byte() {&H0, &H0, &H0, &H0, &H0, &H0, _
             &H90, &H0, &H41})
            WriteData(REGISTER_IR_SENSITIVITY_2, 2, New Byte() {&H40, &H0})

            WriteData(REGISTER_IR_MODE, CByte(mode))


        End Sub

        ''' <summary>
        ''' Disable the IR sensor
        ''' </summary>
        Private Sub DisableIR()
            mWiimoteState.IRState.Mode = IRMode.Off

            ClearReport()
            mBuff(0) = CByte(OutputReport.IR)
            mBuff(1) = GetRumbleBit()
            WriteReport()

            ClearReport()
            mBuff(0) = CByte(OutputReport.IR2)
            mBuff(1) = GetRumbleBit()
            WriteReport()
        End Sub

        ''' <summary>
        ''' Initialize the report data buffer
        ''' </summary>
        Private Sub ClearReport()
            Array.Clear(mBuff, 0, REPORT_LENGTH)
        End Sub

        ''' <summary>
        ''' Write a report to the Wiimote
        ''' </summary>
        Private Sub WriteReport()
            If mAltWriteMethod Then
                HIDImports.HidD_SetOutputReport(Me.mHandle.DangerousGetHandle(), mBuff, CUInt(mBuff.Length))
            Else
                mStream.Write(mBuff, 0, REPORT_LENGTH)
            End If

            Thread.Sleep(100)
        End Sub

        ''' <summary>
        ''' Read data or register from Wiimote
        ''' </summary>
        ''' <param name="address">Address to read</param>
        ''' <param name="size">Length to read</param>
        ''' <returns>Data buffer</returns>
        Public Function ReadData(ByVal address As Integer, ByVal size As Short) As Byte()
            ClearReport()

            mReadBuff = New Byte(size - 1) {}

            mBuff(0) = CByte(OutputReport.ReadMemory)
            mBuff(1) = CByte(((address And &HFF000000UI) >> 24) Or GetRumbleBit())
            mBuff(2) = CByte((address And &HFF0000) >> 16)
            mBuff(3) = CByte((address And &HFF00) >> 8)
            mBuff(4) = CByte(address And &HFF)

            mBuff(5) = CByte((size And &HFF00) >> 8)
            mBuff(6) = CByte(size And &HFF)

            WriteReport()

            If Not mReadDone.WaitOne(1000, False) Then
                Throw New Exception("Error reading data from Wiimote...is it connected?")
            End If

            Return mReadBuff
        End Function

        ''' <summary>
        ''' Write a single byte to the Wiimote
        ''' </summary>
        ''' <param name="address">Address to write</param>
        ''' <param name="data">Byte to write</param>
        Public Sub WriteData(ByVal address As Integer, ByVal data As Byte)
            WriteData(address, 1, New Byte() {data})
        End Sub

        ''' <summary>
        ''' Write a byte array to a specified address
        ''' </summary>
        ''' <param name="address">Address to write</param>
        ''' <param name="size">Length of buffer</param>
        ''' <param name="buff">Data buffer</param>

        Public Sub WriteData(ByVal address As Integer, ByVal size As Byte, ByVal buff As Byte())
            ClearReport()

            mBuff(0) = CByte(OutputReport.WriteMemory)
            mBuff(1) = CByte(((address And &HFF000000UI) >> 24) Or GetRumbleBit())
            mBuff(2) = CByte((address And &HFF0000) >> 16)
            mBuff(3) = CByte((address And &HFF00) >> 8)
            mBuff(4) = CByte(address And &HFF)
            mBuff(5) = size
            Array.Copy(buff, 0, mBuff, 6, size)

            WriteReport()

            Thread.Sleep(100)
        End Sub

        ''' <summary>
        ''' Current Wiimote state
        ''' </summary>
        Public ReadOnly Property WiimoteState() As WiimoteState
            Get
                Return mWiimoteState
            End Get
        End Property

        '''<summary>
        ''' Force the PC to write to the Wiimote in an alternate way.  This should no longer be used.
        '''</summary>
        <Obsolete("The Connect method will determine the proper write method at runtime.  Only set this if you really want to force it.")> _
        Public Property AltWriteMethod() As Boolean
            Get
                Return mAltWriteMethod
            End Get
            Set(ByVal value As Boolean)
                mAltWriteMethod = value
            End Set
        End Property

#Region "IDisposable Members"

        ''' <summary>
        ''' Dispose Wiimote
        ''' </summary>
        Public Sub Dispose() Implements IDisposable.Dispose
            Dispose(True)
            GC.SuppressFinalize(Me)
        End Sub

        ''' <summary>
        ''' Dispose wiimote
        ''' </summary>
        ''' <param name="disposing">Disposing?</param>
        Protected Overridable Sub Dispose(ByVal disposing As Boolean)
            ' close up our handles
            If disposing Then
                Disconnect()
            End If
        End Sub
        Private Shared Function InlineAssignHelper(Of T)(ByRef target As T, ByVal value As T) As T
            target = value
            Return value
        End Function
#End Region
    End Class
End Namespace
